import os
from typing import Optional

from meiga import Failure, Success

from lume.config import SetupConfig
from lume.src.domain.services.logger import INFO, WARNING, Logger
from lume.src.domain.services.setup_service import SetupService
from lume.src.infrastructure.services.setup.setup_errors import (
    ItemTypeNotSupportedError,
)
from lume.src.infrastructure.services.setup.setup_item_bucket import SetupItemBucket
from lume.src.infrastructure.services.setup.setup_item_file import SetupItemFile


class BucketSetupService(SetupService):
    def __init__(self, setup_config: Optional[SetupConfig], logger: Logger):
        if not setup_config:
            return

        self.setup_config = setup_config
        self.logger = logger
        self.set_base_path(setup_config.output)
        self.item_factory = {
            "file": SetupItemFile(setup_config.output),
            "bucket": SetupItemBucket(setup_config.output),
        }

    def execute(self):
        if not self.setup_config:
            self.logger.log(WARNING, "Empty config for setup")
            return

        self.logger.log(INFO, f"setup: output: {self.setup_config.output}")

        for key, value in self.setup_config.deps.items():

            self.logger.log(INFO, f"{key} : {value}")
            if value.type not in self.item_factory.keys():
                return Failure(
                    ItemTypeNotSupportedError(value.type, self.item_factory.keys())
                )

            result = self.item_factory[value.type].run(key, value, self.logger)
            if result.is_success:
                self.logger.log(INFO, f"{key}: Download successfully")
            else:
                return result

        return Success()

    @staticmethod
    def set_base_path(path):
        if not os.path.exists(path):
            os.makedirs(path)



from abc import ABC, abstractmethod

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
COMMAND = 60
ENVAR = 70
ENVAR_WARNING = 80
WAITING = 90
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0
HIGHLIGHT = 100
GLOBAL = 110


class Logger(ABC):
    @abstractmethod
    def log(self, logging_level: int, message: str):
        return NotImplementedError



from typing import Dict, List, Optional

from pydantic import BaseModel

from lume.config.check_os_list_or_str_item import check_os_list_or_str_item
from lume.config.get_envs import get_envs


class UninstallConfig(BaseModel):
    run: List[str]
    cwd: Optional[str] = None
    envs: Dict[str, str] = dict()
    overwrote_envs: List[str] = list()

    def add_shared_env(self, shared_envs: Dict[str, str]):
        if shared_envs and self.envs:
            self.overwrote_envs = list(
                set(shared_envs.keys()).intersection(set(self.envs))
            )

    @staticmethod
    def from_dict(kdict):
        run = check_os_list_or_str_item(kdict, "run")
        envs = get_envs(kdict)
        return UninstallConfig(run=run if run else [], cwd=kdict.get("cwd"), envs=envs)



from .config import Config
from .dependency_config import DependencyConfig
from .install_config import InstallConfig
from .setup_config import SetupConfig
from .step_config import StepConfig
from .uninstall_config import UninstallConfig

__all__ = [
    "Config",
    "DependencyConfig",
    "InstallConfig",
    "UninstallConfig",
    "SetupConfig",
    "StepConfig",
]



import os
from zipfile import BadZipFile

from google.api_core.exceptions import Forbidden, NotFound
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import storage
from meiga import Failure, Result, Success

from lume.config import DependencyConfig
from lume.src.domain.services.logger import INFO, Logger
from lume.src.infrastructure.services.setup.setup_errors import (
    BadZipFileError,
    BlobNotFoundError,
    CrendentialsEnvError,
)
from lume.src.infrastructure.services.setup.setup_item import SetupItem
from lume.src.infrastructure.services.setup.setup_utils import unzip_file


class SetupItemBucket(SetupItem):
    def run(
        self, name: str, dependency_config: DependencyConfig, logger: Logger
    ) -> Result:
        dependency_path = os.path.join(self.base_path, name)
        if not os.path.exists(dependency_path):
            os.makedirs(dependency_path)
        elif dependency_config.overwrite:
            logger.log(
                INFO,
                f"{self.__class__.__name__} - dependency {name} already exists. Overwriting...",
            )
        else:
            logger.log(
                INFO,
                f"{self.__class__.__name__} - dependency {name} already exists. Skipping...",
            )
            return Success()

        if dependency_config.auth_required:
            credentials_path = os.environ.get(dependency_config.credentials_env)
            if credentials_path is None or not os.path.exists(credentials_path):
                try:
                    storage_client = storage.Client()
                except DefaultCredentialsError:
                    return Failure(
                        CrendentialsEnvError(dependency_config.credentials_env)
                    )
            else:
                storage_client = storage.Client.from_service_account_json(
                    credentials_path
                )
        else:
            storage_client = storage.Client()

        try:
            self.__download_bucket(
                storage_client, dependency_path, dependency_config.url
            )
        except NotFound:
            return Failure(BlobNotFoundError(dependency_config.url))
        except Forbidden:
            return Failure(CrendentialsEnvError(dependency_config.credentials_env))

        if dependency_config.unzip:
            try:
                unzip_file(dependency_path, dependency_config.url)
            except BadZipFile:
                return Failure(BadZipFileError(name))

        return Success()

    @staticmethod
    def __download_bucket(storage_client: storage.Client, dst: str, url: str):
        filepath_parts = url.split("/")
        filename = filepath_parts[-1]
        bucket_name = "/".join(filepath_parts[2:-1])
        bucket = storage_client.bucket(bucket_name)
        blob = bucket.blob(filename)
        blob.download_to_filename(os.path.join(dst, filename))



from lume.src.infrastructure.services.logger.colors import Colors


def get_colored_command_message(command, cwd, step, prefix=None) -> str:
    message = (
        f"{Colors.OKBLUE}{step}{Colors.ENDC} {Colors.BOLD}>> {command}{Colors.ENDC}"
        if not cwd
        else f"{Colors.OKBLUE}{step}{Colors.ENDC} {Colors.HEADER}[cwd={cwd}]{Colors.ENDC} {Colors.BOLD}>> {command}{Colors.ENDC}"
    )
    if prefix:
        message = f"{Colors.WARNING}{prefix}{Colors.ENDC} | {message}"
    return message



def check_list_or_str_item(kdict, key, required=False, suffix=""):
    kvalue = kdict.get(key)
    if not kvalue and required:
        raise TypeError(f"StepConfig must contains {key}{suffix} variable")

    if isinstance(kvalue, str):
        value = [kvalue]
    elif isinstance(kvalue, list):
        value = kvalue
    else:
        if required:
            raise TypeError(
                f"StepConfig must contains {key}{suffix} variable (Only list and str is supported)"
            )
        else:
            value = None
    return value



from subprocess import PIPE, Popen

from meiga import Error, Result, isFailure, isSuccess

from lume.src.domain.services.executor_service import (
    DEFAULT_EXECUTOR_LOG_FILENAME,
    ExecutorService,
)
from lume.src.domain.services.logger import ERROR, INFO, WARNING, Logger


class PopenExecutorService(ExecutorService):
    def __init__(self, logger: Logger):
        self.logger = logger

    def execute(
        self, command: str, cwd: str, log_filename: str = DEFAULT_EXECUTOR_LOG_FILENAME
    ) -> Result[bool, Error]:

        if not cwd:
            cwd = "."

        process = Popen(command, stdout=PIPE, stderr=PIPE, cwd=cwd, shell=True)

        while True:
            output = process.stdout.readline()
            return_code = process.poll()
            if return_code is not None:
                break
            if output:
                log_output = output.rstrip().decode("utf-8")
                if log_output != "":
                    self.logger.log(INFO, f"{log_output}")

        for output_err in iter(process.stderr.readline, b""):
            if output_err:
                logging_level = ERROR if return_code != 0 else WARNING
                log_output = output_err.rstrip().decode("utf-8")
                if log_output != "":
                    self.logger.log(logging_level, f"{log_output}")

        if return_code != 0:
            return isFailure

        return isSuccess



from abc import ABC, abstractmethod

from meiga import Error, NotImplementedMethodError, Result

DEFAULT_EXECUTOR_LOG_FILENAME = "lume.executor.log"


class ExecutorService(ABC):
    @abstractmethod
    def execute(self, command: str, cwd: str, log_filename: str) -> Result[bool, Error]:
        return NotImplementedMethodError



import os
import shutil
import zipfile


def unzip_file(dst: str, url: str):
    file_name = url.split("/")[-1]
    path_zipfile = os.path.join(dst, file_name)
    zip_ref = zipfile.ZipFile(path_zipfile, "r")
    zip_ref.extractall(dst)
    zip_ref.close()
    os.remove(path_zipfile)

    if contain_one_folder(dst):
        content = os.listdir(dst)
        name_path = content[0]
        move_content(os.path.join(dst, name_path), dst)
        shutil.rmtree(os.path.join(dst, name_path))


def contain_one_folder(path):
    content = os.listdir(path)
    if len(content) == 1 and os.path.isdir(os.path.join(path, content[0])):
        return True
    return False


def move_content(unzip_directory_path, dst):
    unzip_directory_path_content = os.listdir(unzip_directory_path)
    for item_name in unzip_directory_path_content:
        shutil.move(os.path.join(unzip_directory_path, item_name), dst)



import os
from typing import Dict, List

import yaml
from meiga import BoolResult, Failure, isSuccess

from lume.config.get_envs import get_envs
from lume.config.install_config import InstallConfig
from lume.config.required_env_error import RequiredEnvError
from lume.config.setup_config import SetupConfig
from lume.config.step_config import StepConfig
from lume.config.uninstall_config import UninstallConfig


class Config:
    def __init__(self, lume_dict: Dict):
        self.name = lume_dict.get("name")
        self.settings = {
            "show_exit_code": lume_dict.get("settings", {}).get("show_exit_code", False)
        }

        self.required_env = lume_dict.get("required_env")
        self.shared_envs = get_envs(lume_dict)
        self._set_install_step(lume_dict)
        self._set_uninstall_step(lume_dict)
        self.strict_mode = True
        self.steps = {}
        for step_name, step in lume_dict["steps"].items():
            if step_name == "setup":
                self.steps[step_name] = SetupConfig(**step)
            else:
                self.steps[step_name] = StepConfig.from_dict(step)
                self.steps[step_name].add_shared_env(self.shared_envs)

        self.add_other_steps(lume_dict)

    def update_strict_mode(self, strict_mode: bool):
        self.strict_mode = strict_mode

    def _set_install_step(self, yaml_dict: dict):
        if yaml_dict.get("install"):
            self.install = InstallConfig.from_dict(yaml_dict.get("install"))
        else:
            self.install = InstallConfig(run=[])
        self.install.add_shared_env(self.shared_envs)

    def _set_uninstall_step(self, yaml_dict: dict):
        if yaml_dict.get("uninstall"):
            self.uninstall = UninstallConfig.from_dict(yaml_dict.get("uninstall"))
        else:
            self.uninstall = UninstallConfig(run=[])
        self.uninstall.add_shared_env(self.shared_envs)

    def check_requirements(self) -> BoolResult:
        if self.required_env and self.strict_mode:
            unmeet_required_env_messages = dict()
            for env, description in self.required_env.items():
                if env not in os.environ:
                    unmeet_required_env_messages[env] = description
            if len(unmeet_required_env_messages) > 0:
                return Failure(RequiredEnvError(unmeet_required_env_messages))
        return isSuccess

    def get_steps(self) -> List[str]:
        return list(self.steps.keys())

    def get_commands(self) -> List[str]:
        commands = []
        commands += self.get_steps()
        if self.install:
            commands.append("install")
        if self.uninstall:
            commands.append("uninstall")
        return commands

    def add_other_steps(self, yaml_dict):
        other_steps = yaml_dict.get("other_steps", dict())
        for key, filename in other_steps.items():
            with open(filename) as file:
                yaml_dict = yaml.load(file, Loader=yaml.FullLoader)
                other_shared_envs = get_envs(yaml_dict)
                self.shared_envs.update(other_shared_envs)
                for step_name, step in yaml_dict["steps"].items():
                    step_name = f"{key}:{step_name}"
                    self.steps[step_name] = StepConfig.from_dict(step)
                    self.steps[step_name].add_shared_env(self.shared_envs)



import os
from zipfile import BadZipFile

import requests
from meiga import Failure, Result, Success
from requests.auth import HTTPBasicAuth

from lume.config import DependencyConfig
from lume.src.domain.services.logger import INFO, Logger
from lume.src.infrastructure.services.setup.setup_errors import (
    BadZipFileError,
    CrendentialsEnvError,
)
from lume.src.infrastructure.services.setup.setup_item import SetupItem
from lume.src.infrastructure.services.setup.setup_utils import unzip_file


class SetupItemFile(SetupItem):
    def run(
        self, name: str, dependency_config: DependencyConfig, logger: Logger
    ) -> Result:
        dependency_path = os.path.join(self.base_path, name)
        if os.path.exists(dependency_path):
            logger.log(
                INFO, f"{self.__class__.__name__} - dependency {name} already exists"
            )
            return Success()
        os.makedirs(dependency_path, exist_ok=True)

        if not dependency_config.auth_required:
            auth = None
        else:
            credentials_var = os.environ.get(dependency_config.credentials_env)
            if credentials_var is None:
                return Failure(CrendentialsEnvError(dependency_config.credentials_env))
            username, password = credentials_var.split(":")
            auth = HTTPBasicAuth(username, password)

        self.__download_file(dependency_path, dependency_config.url, auth)

        if dependency_config.unzip:
            try:
                unzip_file(dependency_path, dependency_config.url)
            except BadZipFile:
                return Failure(BadZipFileError(name))

        return Success()

    @staticmethod
    def __download_file(dst: str, url: str, auth: HTTPBasicAuth = None):
        r = requests.get(url, auth=auth, stream=True)
        dst_filename = os.path.split(url)[-1]
        with open(os.path.join(dst, dst_filename), "wb") as f:
            f.write(r.content)



import os

import yaml
from yaml.parser import ParserError


def get_envs(yaml_dict: dict) -> dict:
    envs = get_envs_from_dict(yaml_dict)
    envs_from_file = get_envs_from_file(yaml_dict)
    envs.update(envs_from_file)
    return envs


def read_env_from_file(filename):
    if not filename or not os.path.isfile(filename):
        return {}
    try:
        with open(filename) as file:
            envs = yaml.load(file, Loader=yaml.FullLoader)
            return envs
    except ParserError:
        return {}


def get_envs_from_dict(yaml_dict: dict):
    envs = yaml_dict.get(
        "env", dict()
    )  # official (it matches with GitHub actions syntax)
    if not envs:
        envs = yaml_dict.get(
            "envs", dict()
        )  # fallback to be compatible with old versions of lume.yml
    return envs


def get_envs_from_file(yaml_dict: dict):
    envs = read_env_from_file(yaml_dict.get("env_file"))  # official
    if not envs:
        envs = read_env_from_file(
            yaml_dict.get("envs_file")
        )  # fallback to be compatible with old versions of lume.yml
    return envs
